package com.googlecode.eventio.core;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

import com.googlecode.eventio.Reaction;
import com.googlecode.eventio.core.UniqueMatchGenerator.FilterEventHolder;
import com.googlecode.eventio.core.impl.EventDispatcherImpl;

/**
 * The ResourceAccessManager uses a dedicated thread pool to turn
 * {@link BlockingResource}s into {@link Resource}s: it lets it own threads
 * block until data is available, and uses an {@link EventDispatcherImpl} to execute
 * reactions.
 * 
 * @author mofleury
 * 
 */
public class ResourceAccessManager<FILTER, EVENT> {

	/**
	 * Event dispatcher that will execute callbacks
	 */
	private final EventDispatcherImpl<FILTER, EVENT> eventDispatcher;

	private final UniqueMatchGenerator<FILTER, EVENT> matchGenerator;

	/**
	 * Executor used to access blosking resources
	 */
	private final ExecutorService executor;

	public ResourceAccessManager(EventDispatcherImpl<FILTER, EVENT> eventDispatcher,
			UniqueMatchGenerator<FILTER, EVENT> matchGenerator) {
		this.eventDispatcher = eventDispatcher;
		this.matchGenerator = matchGenerator;
		executor = Executors.newSingleThreadExecutor();
	}

	public ResourceAccessManager(EventDispatcherImpl<FILTER, EVENT> eventDispatcher,
			UniqueMatchGenerator<FILTER, EVENT> matchGenerator, int poolSize,
			ThreadFactory threadFactory) {
		this.eventDispatcher = eventDispatcher;
		this.matchGenerator = matchGenerator;
		executor = Executors.newFixedThreadPool(poolSize, threadFactory);
	}

	public ExecutorService getExecutor() {
		return executor;
	}

	public <INPUT, OUTPUT> Resource<INPUT, OUTPUT> manageResource(
			final BlockingResource<INPUT, OUTPUT> blockingResource) {
		return new Resource<INPUT, OUTPUT>() {
			@Override
			public void get(final INPUT input, final Reaction<Async<OUTPUT>> reaction) {

				final FilterEventHolder<FILTER, EVENT> filterEventHolder = matchGenerator
						.generateUniqueMatch();

				eventDispatcher.once((EVENT) filterEventHolder.filter, reaction);

				executor.execute(new Runnable() {
					@Override
					public void run() {
						try {
							final OUTPUT out = blockingResource.doGet(input);
							eventDispatcher.emit(filterEventHolder.event, new Async<OUTPUT>(out));
						} catch (Throwable t) {

							eventDispatcher.emit(filterEventHolder.event, new Async<OUTPUT>(t));
						}
					}
				});
			};
		};
	}
}
