package com.googlecode.eventio.core.impl;

import java.util.ArrayList;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import com.googlecode.eventio.EventDispatcher;
import com.googlecode.eventio.Reaction;
import com.googlecode.eventio.core.EventMatcher;
import com.googlecode.eventio.core.FailureHandler;
import com.googlecode.eventio.core.impl.util.monitoring.QueueMonitor;

/**
 * The {@link EventDispatcherImpl} controls event flow: it supports event
 * subscription and notification, and calls the registered reactions whenever a
 * event is triggered. The Dispatcher executes all reactions in a dedicated
 * {@link Thread}, which can be parameterized by using a custom
 * {@link ThreadFactory}.
 * 
 * @author mofleury
 * 
 */
public class EventDispatcherImpl<FILTER, EVENT> implements EventDispatcher {

	private final EventMatcher<FILTER, EVENT> reactionRegistry;

	private final ExecutorService executor;

	private final FailureHandler<EVENT> failureHandler = new DefaultFailureHandler<EVENT>();

	public EventDispatcherImpl(EventMatcher<FILTER, EVENT> reactionRegistry) {
		this(Executors.defaultThreadFactory(), reactionRegistry);
	}

	public EventDispatcherImpl(ThreadFactory threadFactory,
			EventMatcher<FILTER, EVENT> reactionRegistry) {
		// TODO use builder pattern to avoid overloaded constructors
		super();
		BlockingQueue<Runnable> queue = new LinkedBlockingQueue<Runnable>();
		new QueueMonitor(new LinkedBlockingQueue<Runnable>(), "eventDispatcher@"
				+ System.identityHashCode(this));

		this.executor = new ThreadPoolExecutor(1, 1, 0L, TimeUnit.MILLISECONDS,//
				queue, threadFactory);
		this.reactionRegistry = reactionRegistry;
	}

	/**
	 * awaits the termination of the event dispatcher executor.
	 * 
	 * @param timeOut
	 * @param timeUnit
	 * @throws InterruptedException
	 */
	public void awaitTermination(int timeOut, TimeUnit timeUnit) throws InterruptedException {
		executor.awaitTermination(timeOut, timeUnit);
	}

	/**
	 * Emit an event, i.e. trigger all reactions, using the provided message as
	 * reaction input
	 * 
	 * @param event
	 *            the event to trigger
	 * @param message
	 *            the message to be reacted upon
	 */
	@Override
	public <T> void emit(final Object event, final T message) {

		executor.execute(new Runnable() {
			@Override
			public void run() {
				for (final Reaction<T> reaction : new ArrayList<Reaction<T>>(reactionRegistry
						.<T> findMatches((EVENT) event))) {
					try {
						reaction.react(message);
					} catch (Throwable t) {
						failureHandler.handleFailure((EVENT) event, message, reaction, t);
					}
				}
			}
		});
	}

	/**
	 * Subscribes this reaction to the given event. The reaction will be invoked
	 * whenever the event is triggered. Subscriptions can be canceled using
	 * {@link #remove(Object, Reaction)}
	 * 
	 * @param event
	 *            the event to subscribe to
	 * @param reaction
	 *            the reaction to be executed when the event arises
	 */
	@Override
	public <T> void on(Object event, Reaction<T> reaction) {
		reactionRegistry.register((FILTER) event, reaction);
	}

	/**
	 * Same as {@link #on(Object, Reaction)}, except that the reaction will be
	 * invoked only once.
	 * 
	 * @param event
	 *            the event to subscribe to
	 * @param reaction
	 *            the reaction to be executed when the event arises
	 */
	@Override
	public <T> void once(final Object event, final Reaction<T> reaction) {
		Reaction<T> selfRemovingReaction = new Reaction<T>() {
			@Override
			public void react(T input) {
				remove((FILTER) event, this);
				reaction.react(input);
			};
		};

		on(event, selfRemovingReaction);
	}

	/**
	 * Removes reaction, if any. Reaction comparison is done by reference. If
	 * the reaction is present more than once, only the most recently added
	 * instance will be removed
	 * 
	 * @param event
	 *            the event to remove reactions from
	 * @param reaction
	 *            the reaction to be removed
	 */
	public <T> void remove(FILTER event, Reaction<T> reaction) {
		reactionRegistry.remove(event, reaction);
	}

	/**
	 * removes all reactions that correspond to this event
	 * 
	 * @param event
	 */
	public void removeAll(FILTER event) {
		reactionRegistry.removeAll(event);
	}

}
